-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Expo: Profiler")
local dewdrop = AceLibrary("Dewdrop-2.0")

L:RegisterTranslations("enUS", function() return {
	["Profiler"] = true,
	["Configure..."] = true,
	['New Generic Profiler...'] = true,
	['New AddOn Profiler...'] = true,
	['New Library Profiler...'] = true,
	['New Class Profiler...'] = true,
	['Delete Profiler...'] = true,
	['Sort by time'] = true,
	['Sort by time/call'] = true,
	['Sort by memory usage'] = true,
	['Sort by name'] = true,
	['Sort by calls'] = true,
	['No Profilers Currently Tracked'] = true,
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

local db
local buffer
local table_remove = table.remove
local prefuncs = {}  -- This stores all functions that are wrapped, allowing Profiler to restore the original when necessary.
local tree = {}
local treeMemories = {}
local treeTimes = {}
local memories = {}
local times = {}
local timepers = {}
local calls = {}
local headings = {}
local shorts = {}
local sorts = {
	time = function(alpha, bravo)
		if times[alpha] ~= times[bravo] then
			return times[alpha] > times[bravo]
		elseif memories[alpha] ~= memories[bravo] then	
			return memories[alpha] > memories[bravo]
		elseif calls[alpha] ~= calls[bravo] then
			return calls[alpha] > calls[bravo]
		else
			return alpha:lower() > bravo:lower()
		end
	end,
	timeper = function(alpha, bravo)
		if timepers[alpha] ~= timepers[bravo] then
			return timepers[alpha] > timepers[bravo]
		elseif memories[alpha] ~= memories[bravo] then	
			return memories[alpha] > memories[bravo]
		elseif calls[alpha] ~= calls[bravo] then
			return calls[alpha] > calls[bravo]
		else
			return alpha:lower() > bravo:lower()
		end
	end,
	memory = function(alpha, bravo)
		if memories[alpha] ~= memories[bravo] then	
			return memories[alpha] > memories[bravo]
		elseif times[alpha] ~= times[bravo] then
			return times[alpha] > times[bravo]
		elseif calls[alpha] ~= calls[bravo] then
			return calls[alpha] > calls[bravo]
		else
			return alpha:lower() > bravo:lower()
		end
	end,
	calls = function(alpha, bravo)
		if calls[alpha] ~= calls[bravo] then
			return calls[alpha] > calls[bravo]
		elseif times[alpha] ~= times[bravo] then
			return times[alpha] > times[bravo]
		elseif memories[alpha] ~= memories[bravo] then	
			return memories[alpha] > memories[bravo]
		else
			return alpha:lower() > bravo:lower()
		end
	end,
	name = function(alpha, bravo)
		return alpha:lower() < bravo:lower()
	end,
}

local function wrap(value, name, heading, short)
	if type(value) == "function" then
		local oldFunction = value
		prefuncs[name] = value
		memories[name] = 0
		times[name] = 0
		timepers[name] = 0
		calls[name] = 0
		headings[name] = heading
		shorts[name] = short or name
		local function func(pos, t, mem, ...)
			mem, t = collectgarbage('count') - mem, GetTime() - t
			if pos > 0 then
				treeMemories[pos] = treeMemories[pos] + mem
				treeTimes[pos] = treeTimes[pos] + t
			end
			local otherMem = table_remove(treeMemories)
			if mem - otherMem > 0 then
				memories[name] = memories[name] + mem - otherMem
			end
			times[name] = times[name] + t - table_remove(treeTimes)
			timepers[name] = times[name] / calls[name]
			table_remove(tree)
			return ...
		end
		return function(...)
			local pos = #tree
			tree[#tree+1] = name
			treeMemories[#treeMemories+1] = 0
			treeTimes[#treeTimes+1] = 0
			calls[name] = calls[name] + 1
			local t, mem = GetTime(), collectgarbage('count')
			return func(pos, t, mem, oldFunction(...))
		end
	elseif type(value) == "table" then
		for k, v in pairs(value) do
			if type(k) == "string" and type(v) == "function" and v ~= wrap then
				value[k] = wrap(v, name .. '.' .. k,heading,':'..k)
			end
		end
		return value
	end
end

local function unwrap(name,value)
	if type(value) == "function" then
		local ret = prefuncs[name]
		prefuncs[name] = nil
		memories[name] = nil
		times[name] = nil
		timepers[name] = nil
		calls[name] = nil
		--DEFAULT_CHAT_FRAME:AddMessage("unwrapped "..name)
		return ret
	elseif type(value) == "table" then
		for k, v in pairs(value) do
			if type(k) == "string" and type(v) == "function" and v ~= wrap then
				value[k] = unwrap(name .. '.' .. k,v)
			end
		end
		--DEFAULT_CHAT_FRAME:AddMessage("unwrapped "..name)
		return value
	end
end

local function getdeep(str)
	local pre_ = _
	_ = nil
	RunScript("_ = "..str)
	local ret = _
	_ = pre_
	return ret
end

-----------------------------------------------------------------------
-- Module declaration
-----------------------------------------------------------------------

local mod = Expo:NewModule("Profiler")
mod.revision = tonumber(("$Revision: 32998 $"):sub(12, -3)) or 1
mod.version = 1

-----------------------------------------------------------------------
-- Module initialization
-----------------------------------------------------------------------

function mod:OnModuleRegister()
end

function mod:OnModuleEnable()
	-- set up a database to store event filters, pause/resume status, and max buffer size.
	self.db = Expo:AcquireDBNamespace("Profiler")
	db = self.db.account
	db.wraps = db.wraps or {}
	db.display = db.display or {}
	db.virtual = db.virtual or {}
	db.currentSort = db.currentSort or 'time'
	-- Wrap the functions and tables.
	for i,curWrap in ipairs(db.wraps) do
		local value = getdeep(curWrap)
		if type(value) == 'function' or type(value) == 'table' then
			wrap(value,curWrap,GetTime())
		end
	end
	self.core:UpdateTooltip()
end

function mod:OnModuleDisable()
	for i,wrap in ipairs(db.wraps) do
		local preunwrap = tunwrap
		tunwrap = unwrap
		RunScript(wrap.." = tunwrap('"..wrap.."',"..wrap..")")
		tunwrap = preunwrap
	end
end

-----------------------------------------------------------------------
-- Module implementation
-----------------------------------------------------------------------

function mod:GetRevision() return self.revision end
function mod:GetDisplayName() return L["Profiler"] end

function mod:TrackNewWrap(newWrap,displayName)
	local value = getdeep(newWrap)
	if type(value) == 'function' or type(value) == 'table' then
		displayName = displayName or newWrap
		wrap(value,newWrap,displayName)
		db.display[newWrap] = displayName
		db.wraps[#db.wraps + 1] = newWrap
	end
end

function mod:UntrackWrap(wrapIndex)
	local name = db.wraps[wrapIndex]
	local preunwrap = tunwrap
	tunwrap = unwrap
	RunScript(name.." = tunwrap('"..name.."',"..name..")")
	tunwrap = preunwrap
	db.display[name] = nil
	table_remove(db.wraps,wrapIndex)
end

function mod:GetNumberOfColumns() return 6 end
function mod:GetData(cat)
	-- add line to add and remove events
	cat:AddLine(
		'text', " ",
		'text2', " ",
		'text3', " ",
		'text4', " ",
		'text5', " ",
		'text6', L["Configure..."],
		'text6R', 0.6,
		'text6G', 0.6,
		'text6B', 1,
		'func',function() self:PopupConfigureMenu() end
	)
	cat:AddLine(
		'text', "Function",
		'text2', "Time",
		'text3', "Memory",
		'text4', "Calls",
		'text5', "Time/Call",
		'text6', "Memory/Call"
	)
	for k,v in pairs(db.display) do
		cat:AddLine(
			'text', v
		)
		local t = self:new()
		for method,hdng in pairs(headings) do
			if hdng == v then
				t[#t+1] = method
			end
		end
		table.sort(t, sorts[db.currentSort])
		for _,method in ipairs(t) do
			if calls[method] > 0 then
				cat:AddLine(
					'text', shorts[method],
					'text2', ("%.3f s"):format(times[method]),
					'text3', ("%.3f KiB"):format(memories[method]),
					'text4', ("%dx"):format(calls[method]),
					'text5', ("%.3f ms"):format(times[method] / calls[method] * 1000),
					'text6', ("%.3f KiB"):format(memories[method] / calls[method]),
					'func', function() end,
					'indentation',10
				)
			end
		end
		self:del(t)
	end
end

function mod:PopupConfigureMenu()
	if not dewdrop:IsOpen("Expo_Profiler") then
		dewdrop:Close()
		dewdrop:Open("Expo_Profiler",'children',function(level,v,v_1,v_2,v_3,v_4)
				if level == 1 then
					dewdrop:AddLine('text',L['New Generic Profiler...'],
						'hasArrow',true,
						'hasEditBox',true,
						'editBoxFunc',function(newValue) self:TrackNewWrap(newValue) self.core:UpdateTooltip() end,
						'tooltipTitle','Common Patterns:',
						'tooltipText','AceLibrary("MyClass-1.0").prototype\nAceLibrary("MyAddon-1.0")'
					)
					dewdrop:AddLine('text',L['New AddOn Profiler...'],
						'hasArrow',true,
						'hasEditBox',true,
						'editBoxFunc',function(newValue)
							self:TrackNewWrap(newValue,'AddOn: '..newValue)
							local v = getdeep(newValue)
							if v and type(v.IterateModules) == 'function' then
								for name, module in v:IterateModules() do
									self:TrackNewWrap(newValue..':GetModule("'..name..'")',newValue..' Module: '..name)
									db.virtual[newValue..':GetModule("'..name..'")'] = newValue
								end
							end
							self.core:UpdateTooltip()
						end,
						'tooltipTitle','Common Patterns:',
						'tooltipText','AceLibrary("MyClass-1.0").prototype\nAceLibrary("MyAddon-1.0")'
					)
					dewdrop:AddLine('text',L['New Library Profiler...'],
						'hasArrow',true,
						'hasEditBox',true,
						'editBoxFunc',function(newValue) self:TrackNewWrap('AceLibrary("'..newValue..'")','Library: '..newValue) self.core:UpdateTooltip() end,
						'tooltipTitle','Common Patterns:',
						'tooltipText','AceLibrary("MyClass-1.0").prototype\nAceLibrary("MyAddon-1.0")'
					)
					dewdrop:AddLine('text',L['New Class Profiler...'],
						'hasArrow',true,
						'hasEditBox',true,
						'editBoxFunc',function(newValue) self:TrackNewWrap('AceLibrary("'..newValue..'").prototype','Class: '..newValue) self.core:UpdateTooltip() end,
						'tooltipTitle','Common Patterns:',
						'tooltipText','AceLibrary("MyClass-1.0").prototype\nAceLibrary("MyAddon-1.0")'
					)
					dewdrop:AddLine('text',L['Delete Profiler...'],
						'hasArrow',true,'value','DeleteWraps')
					dewdrop:AddLine('text','')
					dewdrop:AddLine('text',L['Sort by time'],
						'checked',db.currentSort == 'time',
						'func',function() db.currentSort = 'time' end)
					dewdrop:AddLine('text',L['Sort by time/call'],
						'checked',db.currentSort == 'timeper',
						'func',function() db.currentSort = 'timeper' end)
					dewdrop:AddLine('text',L['Sort by memory usage'],
						'checked',db.currentSort == 'memory',
						'func',function() db.currentSort = 'memory' end)
					dewdrop:AddLine('text',L['Sort by name'],
						'checked',db.currentSort == 'name',
						'func',function() db.currentSort = 'name' end)
					dewdrop:AddLine('text',L['Sort by calls'],
						'checked',db.currentSort == 'calls',
						'func',function() db.currentSort = 'calls' end)
				elseif level == 2 and v == 'DeleteWraps' then
					local no = true
					for i,wrap in ipairs(db.wraps) do
						if not db.virtual[wrap] then
							dewdrop:AddLine('text',db.display[wrap],
								'func',function()
									local v = getdeep(wrap)
									if v and type(v.IterateModules) == 'function' then
										for name, module in v:IterateModules() do
											unwrap(wrap..':GetModule("'..name..'")',module)
											db.virtual[wrap..':GetModule("'..name..'")'] = nil
											db.display[wrap..':GetModule("'..name..'")'] = nil
											for i,v in ipairs(db.wraps) do if v == wrap..':GetModule("'..name..'")' then
												table_remove(db.wraps,i)
												break
											end end
										end
									end
									self:UntrackWrap(i)
									self.core:UpdateTooltip()
								end)
							no = nil
						end
					end
					if no then
						dewdrop:AddLine('text',L['No Profilers Currently Tracked'])
					end
				end
			end,
			'cursorX',true,'cursorY',true)
	else
		dewdrop:Close()
	end
end
